/*
 *  linux/fs/nfs/dir.c
 *
 *  Copyright (C) 1992  Rick Sladkey
 *
 *  nfs directory handling functions
 *
 * 10 Apr 1996	Added silly rename for unlink	--okir
 * 28 Sep 1996	Improved directory cache --okir
 * May-June 1997 Directory cache improvements by
 *       Douglas Zongker and Greg J. Badros
 *       {dougz,gjb}@u.washington.edu
 */

#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/stat.h>
#include <linux/fcntl.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/sunrpc/types.h>
#include <linux/nfs_fs.h>
#include "profile.h"
#include "dir.h"

#include <asm/segment.h>	/* for fs functions */
#include "cache_failed_lookups.h"

/* #define NFS_INSTRUMENT_RPC_LOOKUPS */

#ifdef KERNEL_2_1
#define RW_COUNT_TYPE unsigned long
#else
#define RW_COUNT_TYPE int
#endif

int c_nfs_ops[i_nfs_profile_last_entry+1];
int nfs_async_lookups_enabled = 1;

/*
 * Head for a dircache entry. Currently still very simple; when
 * the cache grows larger, we will need a LRU list.
 */
struct nfs_dirent {
	dev_t			dev;		/* device number */
	ino_t			ino;		/* inode number */
	u32			cookie;		/* cooke of first entry */
	unsigned short		valid  : 1,	/* data is valid */
				locked : 1;	/* entry locked */
	unsigned int		size;		/* # of entries */
	unsigned long		age;		/* last used */
	unsigned long		mtime;		/* last attr stamp */
	struct wait_queue *	wait;
	struct nfs_entry *	entry;
};

static int nfs_dir_open(struct inode * inode, struct file * file);
static int nfs_dir_read(struct inode *, struct file *, char *, RW_COUNT_TYPE);
static int nfs_readdir(struct inode *, struct file *, void *, filldir_t);
static int nfs_lookup(struct inode *, const char *, int, struct inode **);
static int nfs_create(struct inode *, const char *, int, int, struct inode **);
static int nfs_mkdir(struct inode *, const char *, int, int);
static int nfs_rmdir(struct inode *, const char *, int);
static int nfs_unlink(struct inode *, const char *, int);
static int nfs_symlink(struct inode *, const char *, int, const char *);
static int nfs_link(struct inode *, struct inode *, const char *, int);
static int nfs_mknod(struct inode *, const char *, int, int, int);
static int nfs_rename(struct inode *, const char *, int,
		      struct inode *, const char *, int, int);

static struct file_operations nfs_dir_operations = {
	NULL,			/* lseek - default */
	nfs_dir_read,		/* read - bad */
	NULL,			/* write - bad */
	nfs_readdir,		/* readdir */
	NULL,			/* select - default */
	NULL,			/* ioctl - default */
	NULL,			/* mmap */
	nfs_dir_open,		/* open - revalidate */
	NULL,			/* no special release code */
	NULL			/* fsync */
};

struct inode_operations nfs_dir_inode_operations = {
	&nfs_dir_operations,	/* default directory file-ops */
	nfs_create,		/* create */
	nfs_lookup,		/* lookup */
	nfs_link,		/* link */
	nfs_unlink,		/* unlink */
	nfs_symlink,		/* symlink */
	nfs_mkdir,		/* mkdir */
	nfs_rmdir,		/* rmdir */
	nfs_mknod,		/* mknod */
	nfs_rename,		/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,			/* readpage */
	NULL,			/* writepage */
	NULL,			/* bmap */
	NULL,			/* truncate */
	NULL,			/* permission */
	NULL,			/* smap */
	NULL,			/* updatepage */
	nfs_revalidate,		/* revalidate */
};

static int
nfs_dir_open(struct inode *dir, struct file *file)
{
 	PROFILE_INC_COUNT(i_dir_open);
 
	dfprintk(VFS, "NFS: nfs_dir_open(%x/%ld)\n", dir->i_dev, dir->i_ino);
	return nfs_revalidate_inode(NFS_SERVER(dir), dir);
}

static int
nfs_dir_read(struct inode *inode, struct file *filp, char *buf, RW_COUNT_TYPE count)
{
	return -EISDIR;
}

static struct nfs_dirent	dircache[NFS_MAX_DIRCACHE];

/*
 * We need to do caching of directory entries to prevent an
 * incredible amount of RPC traffic.  Only the most recent open
 * directory is cached.  This seems sufficient for most purposes.
 * Technically, we ought to flush the cache on close but this is
 * not a problem in practice.
 *
 * XXX: Do proper directory caching by stuffing data into the
 * page cache (may require some fiddling for rsize < PAGE_SIZE).
 */

static int
nfs_readdir(struct inode *inode, struct file *filp, void *dirent,
						filldir_t filldir)
{
	static struct wait_queue *readdir_wait = NULL;
	struct wait_queue	**waitp = NULL;
	struct nfs_dirent	*cache, *free;
	struct nfs_entry	*entry;
	unsigned long		age, dead;
	u32			cookie;
	int			ismydir, result;
	int			i, j, index = 0;

 	PROFILE_INC_COUNT(i_readdir);
 
	dfprintk(VFS, "NFS: nfs_readdir(%x/%ld)\n", inode->i_dev, inode->i_ino);
	if (!inode || !S_ISDIR(inode->i_mode)) {
		printk("nfs_readdir: inode is NULL or not a directory\n");
		return -EBADF;
	}

	if ((result = nfs_revalidate_inode(NFS_SERVER(inode), inode)) < 0)
		return result;

	/*
	 * Try to find the entry in the cache
	 */
again:
	if (waitp) {
		interruptible_sleep_on(waitp);
		if (current->signal & ~current->blocked)
			return -ERESTARTSYS;
		waitp = NULL;
	}

	cookie = filp->f_pos;
	entry  = NULL;
	free   = NULL;
	age    = ~(unsigned long) 0;
	dead   = jiffies - NFS_ATTRTIMEO(inode);

	for (i = 0, cache = dircache; i < NFS_MAX_DIRCACHE; i++, cache++) {
		/*
		dprintk("NFS: dircache[%d] valid %d locked %d\n",
					i, cache->valid, cache->locked);
		 */
		ismydir = (cache->dev == inode->i_dev
				&& cache->ino == inode->i_ino);
		if (cache->locked) {
			if (!ismydir || cache->cookie != cookie)
				continue;
			dfprintk(DIRCACHE, "NFS: waiting on dircache entry\n");
			waitp = &cache->wait;
			goto again;
		}

		if (ismydir && cache->mtime != NFS_OLDMTIME(inode))
			cache->valid = 0;

		if (!cache->valid || cache->age < dead) {
			free = cache;
			age  = 0;
		} else if (cache->age < age) {
			free = cache;
			age  = cache->age;
		}

		if (!ismydir || !cache->valid)
			continue;

		if (cache->cookie == cookie && cache->size > 0) {
			entry = cache->entry + (index = 0);
			cache->locked = 1;
			break;
		}
		for (j = 0; j < cache->size; j++) {
			/*
			dprintk("NFS: examing entry %.*s @%d\n",
				(int) cache->entry[j].length,
				cache->entry[j].name,
				cache->entry[j].cookie);
			 */
			if (cache->entry[j].cookie != cookie)
				continue;
			if (j < cache->size - 1) {
				entry = cache->entry + (index = j + 1);
			} else if (cache->entry[j].eof) {
				return 0;
			}
			break;
		}
		if (entry) {
			dfprintk(DIRCACHE, "NFS: found dircache entry %d\n",
						cache - dircache);
			cache->locked = 1;
			break;
		}
	}

	/*
	 * Okay, entry not present in cache, or locked and inaccessible.
	 * Set up the cache entry and attempt a READDIR call.
	 */
	if (entry == NULL) {
		if ((cache = free) == NULL) {
			dfprintk(DIRCACHE, "NFS: dircache contention\n");
			waitp = &readdir_wait;
			goto again;
		}
		dfprintk(DIRCACHE, "NFS: using free dircache entry %d\n",
				free - dircache);
		cache->cookie = cookie;
		cache->locked = 1;
		cache->valid  = 0;
		cache->dev    = inode->i_dev;
		cache->ino    = inode->i_ino;
		if (!cache->entry) {
			cache->entry = (struct nfs_entry *)
						get_free_page(GFP_KERNEL);
			if (!cache->entry) {
				result = -ENOMEM;
				goto done;
			}
		}

		result = nfs_proc_readdir(NFS_SERVER(inode), NFS_FH(inode),
					cookie, PAGE_SIZE, cache->entry);
		if (result <= 0)
			goto done;
		cache->size  = result;
		cache->valid = 1;
		entry = cache->entry + (index = 0);
	}
	cache->mtime = NFS_OLDMTIME(inode);
	cache->age = jiffies;

	/*
	 * Yowza! We have a cache entry...
	 */
	while (index < cache->size) {
		int	nextpos = entry->cookie;

		/*
		dprintk("NFS: filldir(%p, %.*s, %d, %d, %x, eof %x)\n", entry,
				(int) entry->length, entry->name, entry->length,
				(unsigned int) filp->f_pos,
				entry->fileid, entry->eof);
		 */

		if (filldir(dirent, entry->name, entry->length, cookie, entry->fileid) < 0)
			break;
		cookie = nextpos;
		if (nextpos != entry->cookie) {
			printk("nfs_readdir: shouldn't happen!\n");
			break;
		}
		index++;
		entry++;
	}
	filp->f_pos = cookie;
	result = 0;

	/* XXX: May want to kick async readdir-ahead here. Not too hard
	 * to do. */

done:
	dfprintk(DIRCACHE, "NFS: nfs_readdir complete\n");
	cache->locked = 0;
	wake_up(&cache->wait);
	wake_up(&readdir_wait);

	return result;
}

/*
 * Invalidate dircache entries for inode
 */
void
nfs_invalidate_dircache(struct inode *inode)
{
	struct nfs_dirent *cache;
	dev_t		dev = inode->i_dev;
	ino_t		ino = inode->i_ino;
	int		i;

	dfprintk(DIRCACHE, "NFS: invalidate dircache for %x/%ld\n", dev, ino);
	for (i = 0, cache = dircache; i < NFS_MAX_DIRCACHE; i++, cache++) {
		if (!cache->locked && cache->dev == dev && cache->ino == ino)
			cache->valid = 0;	/* brute force */
	}
}

/*
 * Free directory cache memory
 * Called from cleanup_module
 */
void
nfs_free_dircache(void)
{
	struct nfs_dirent *cache;
	int		i;

	dfprintk(DIRCACHE, "NFS: freeing dircache\n");
	for (i = 0, cache = dircache; i < NFS_MAX_DIRCACHE; i++, cache++) {
		cache->valid = 0;
		if (cache->locked) {
			printk("nfs_kfree_cache: locked entry in dircache!\n");
			continue;
		}
		if (cache->entry)
			free_page((unsigned long) cache->entry);
		cache->entry = NULL;
	}
}
 

/*
 * Lookup caching is a big win for performance but this is just
 * a trial to see how well it works on a small scale.
 * For example, bash does a lookup on ".." 13 times for each path
 * element when running pwd.  Yes, hard to believe but true.
 * Try pwd in a filesystem mounted with noac.
 *
 * It trades a little cpu time and memory for a lot of network bandwidth.
 * Since the cache is not hashed yet, it is a good idea not to make it too
 * large because every lookup looks through the entire cache even
 * though most of them will fail.
 *
 * FIXME: The lookup cache should also cache failed lookups. This can
 * be a considerable win on diskless clients.
 * 
 * NO! Failed lookup caching could worsen semantics for file-based pseudo-locks
 * over NFS.  When we do a lookup, we need to know for *sure* that the file
 * does not exist on the server.  Remembering a failed lookup is best left to
 * the user level program --gjb
 */

/* dir.h has the DEBUG_NEWCACHE definition (possibly commented out) */

#ifdef DEBUG_NEWCACHE
#define lc_nfs_printk(x...) printk( ## x)
#else
#define lc_nfs_printk(x...)
#endif


#ifdef NEWCACHE
struct nfs_lookup_cache_entry *nfs_lookup_cache;
#else
struct nfs_lookup_cache_entry nfs_lookup_cache[NFS_LOOKUP_CACHE_SIZE];
#endif

#ifdef NEWCACHE
int *nfs_lookup_cache_buckets;
#endif

/* hash (dir,filename) to a value in 0..(NFS_LOOKUP_CACHE_BUCKETS-1) */

#ifdef NEWCACHE
static void nfs_lookup_cache_check ( void )
{
     int i, j;
     int free = 0;
     int lost = 0;

     for ( i = 0; i < NFS_LOOKUP_CACHE_SIZE; ++i )
	  nfs_lookup_cache[i].bit = -1;

     for ( i = 0; i < NFS_LOOKUP_CACHE_BUCKETS; ++i )
     {
	  j = nfs_lookup_cache_buckets[i];
	  if ( j )
	       lc_nfs_printk ( "NFS scan: bucket %d: ", i );

	  while ( j )
	  {
	       lc_nfs_printk ( " %d", j );
	       if ( nfs_lookup_cache[j].bit != -1 )
	       {
		    lc_nfs_printk ( "NFS error:  index %d in buckets %d and %d\n",
				 j, nfs_lookup_cache[j].bit, i );
	       }
	       nfs_lookup_cache[j].bit = i;
	       j = nfs_lookup_cache[j].next;
	  }
	  if ( nfs_lookup_cache_buckets[i] )
	       lc_nfs_printk ( "\n" );
     }

     lc_nfs_printk ( "NFS: bucket scan complete.\n" );
     
     j = nfs_lookup_cache[0].next;
     while ( j )
     {
	  ++free;
	  if ( nfs_lookup_cache[j].bit >= 0 )
	  {
	       lc_nfs_printk ( "NFS error: index %d on free list and in bucket %d\n",
			    j, nfs_lookup_cache[j].bit );
	  }
	  else if ( nfs_lookup_cache[j].bit == -2 )
	  {
	       lc_nfs_printk ( "NFS error: index %d on free list multiple times.\n", j );
	       break;
	  }
	  nfs_lookup_cache[j].bit = -2;
	  j = nfs_lookup_cache[j].next;
     }

     lc_nfs_printk ( "NFS: free list scan complete: %d free entries.\n", free );

     for ( i = 1; i < NFS_LOOKUP_CACHE_SIZE; ++i )
	  if ( nfs_lookup_cache[i].bit == -1 )
	  {
	       if ( !lost ) 
		    lc_nfs_printk ( "NFS error: lost indices " );
	       lost = 1;
	       lc_nfs_printk ( " %d", i );
	  }
     if ( lost )
	  lc_nfs_printk ( "\n" );

     lc_nfs_printk ( "NFS: lookup cache check complete.\n" );
}
#endif
     
#ifdef NEWCACHE
static int nfs_lookup_cache_hash ( struct inode *dir, const char *filename )
{
     unsigned int i, k;
     k = dir->i_dev + (dir->i_ino<<3);
     i = 0;
     while ( *filename )
     {
	  k += (*filename) << (i++);
	  ++filename;
     }
     return k%NFS_LOOKUP_CACHE_BUCKETS;
}

#endif

/* search for (dir,filename) return a pointer to the entry or NULL. */
/* if (dir,filename) is found, index contains the index of the record,
 * prevptr contains the address of the integer pointing to that record
 * (either the next field of another record, or a hash bucket).
 *
 * if not found, then:
 *    if the hash bucket is empty, index contains the bucket number
 *    and prevptr is NULL
 *    if the hash bucket it not empty, index contains the index of the
 *    last record in the chain, and prevptr is not NULL.
 */

struct nfs_lookup_cache_entry *
nfs_lookup_cache_index(struct inode *dir, const char *filename
#ifdef NEWCACHE
		       , int *index, int **prevptr
#endif
     )
{
#ifndef NEWCACHE
	struct nfs_lookup_cache_entry *entry;
	int i;

	for (i = 0; i < NFS_LOOKUP_CACHE_SIZE; i++) {
		entry = nfs_lookup_cache + i;
		if (entry->dev == dir->i_dev
		    && entry->inode == dir->i_ino
		    && !strncmp(filename, entry->filename, NFS_MAXNAMLEN))
			return entry;
	}
	return NULL;
#else
	struct nfs_lookup_cache_entry *entry;
	int i, j;

	
	j = nfs_lookup_cache_hash ( dir, filename );
	i = nfs_lookup_cache_buckets[j];
	if ( i == 0 )
	{
	     *prevptr = NULL;
	     *index = j;
	     lc_nfs_printk ( "NFS cache_index: dir %p filename \"%s\" empty bucket %d\n",
			  dir, filename, j );
	     return NULL;
	}
	*prevptr = &(nfs_lookup_cache_buckets[j]);
	while ( i )
	{
	     entry = nfs_lookup_cache + i;
	     *index = i;
	     if ( entry->dev == dir->i_dev &&
		  entry->inode == dir->i_ino &&
		  !strncmp ( filename, entry->filename, NFS_MAXNAMLEN ) )
	     {
		  lc_nfs_printk ( "NFS cache_index: dir %p filename \"%s\" index %d bucket %d\n",
			       dir, filename, *index, j );
		  return entry;
	     }
	     i = entry->next;
	     *prevptr = &(entry->next);
	}
	lc_nfs_printk ( "NFS cache_index: dir %p filename \"%s\" not in chain %d\n",
		     dir, filename, j );
	return NULL;
#endif
}

/* do a lookup on (dir,filename) from the cache, fill in fhandle and fattr,
   return 1 if successful, 0 otherwise. */

void nfs_lookup_cache_allocate ( void )
{
#ifdef NEWCACHE
     int i;
     
     lc_nfs_printk ( "NFS: allocating lookup cache.\n" );
     lc_nfs_printk ( "NFS: a cache entry takes %d bytes.\n",
		  sizeof ( struct nfs_lookup_cache_entry ) );
     nfs_lookup_cache = (struct nfs_lookup_cache_entry *)kmalloc ( NFS_LOOKUP_CACHE_SIZE * sizeof ( struct nfs_lookup_cache_entry ), GFP_KERNEL );
     if (nfs_lookup_cache == NULL) {
	printk("NFS: nfs_lookup_cache_allocate: failed kmalloc\n");
	return;
     }
     memset ( nfs_lookup_cache, 0, NFS_LOOKUP_CACHE_SIZE * sizeof ( struct nfs_lookup_cache_entry ) );
     /* all buckets start out empty. */
     nfs_lookup_cache_buckets = (int *)kmalloc ( sizeof(int)*NFS_LOOKUP_CACHE_BUCKETS, GFP_KERNEL );
     if ( nfs_lookup_cache_buckets == NULL )
     {
	  printk ( "NFS: nfs_lookup_cache_allocate: failed kmalloc (buckets)\n" );
	  return;
     }
     memset ( nfs_lookup_cache_buckets, 0,
	      sizeof ( int ) * NFS_LOOKUP_CACHE_BUCKETS );
     /* initialize the free list. */
     for ( i = 0; i < NFS_LOOKUP_CACHE_SIZE; ++i )
	  nfs_lookup_cache[i].next = (i+1)%NFS_LOOKUP_CACHE_SIZE;
#else
     lc_nfs_printk ( "NFS: using old cache\n" );
#endif
}

void nfs_lookup_cache_free ( void )
{
#ifdef NEWCACHE
     if ( nfs_lookup_cache != NULL )
     {
	  lc_nfs_printk ( "NFS: freeing lookup cache.\n" );
	  kfree ( nfs_lookup_cache );
	  nfs_lookup_cache = NULL;
     }
     if ( nfs_lookup_cache_buckets != NULL )
     {
	  kfree ( nfs_lookup_cache_buckets );
	  nfs_lookup_cache_buckets = NULL;
     }
     
#endif
	  
}

static int nfs_lookup_cache_lookup(struct inode *dir, const char *filename,
				   struct nfs_fh *fhandle,
				   struct nfs_fattr *fattr)
{
#ifndef NEWCACHE
	static int nfs_lookup_cache_in_use = 0;

	struct nfs_lookup_cache_entry *entry;

	dfprintk(LOOKUPCACHE, "NFS: lookup_cache_lookup(%x/%ld, %s)\n",
				dir->i_dev, dir->i_ino, filename);
 	PROFILE_INC_COUNT(i_lookup_cache_index);
 	
	if (!nfs_lookup_cache_in_use) {
		memset(nfs_lookup_cache, 0, sizeof(nfs_lookup_cache));
		nfs_lookup_cache_in_use = 1;
	}
	if ((entry = nfs_lookup_cache_index(dir, filename))) {
		if (jiffies > entry->expiration_date) {
			entry->dev = 0;
			PROFILE_INC_COUNT(i_lookup_cache_index_expired);
			return 0;
		}
		*fhandle = entry->fhandle;
		*fattr = entry->fattr;
		return 1;
	}
	return 0;
#else
	struct nfs_lookup_cache_entry *entry;
	int index, *prevptr;

	if ( strlen ( filename ) > NFSLOOKUPCACHENAMELEN )
	     return 0;

#ifdef DEBUG_NEWCACHE
	if ( strcmp ( filename, "dump" ) == 0 )
	{
	     nfs_lookup_cache_check();
	     return 0;
	}
#endif
	
	lc_nfs_printk ( "NFS cache_lookup: dir %p filename \"%s\"\n",
		     dir, filename );
	
	if ( ( entry = nfs_lookup_cache_index ( dir, filename,
					      &index, &prevptr ) ) )
	{
	     lc_nfs_printk ( "NFS cache_lookup: found entry %d\n", index );
	     if ( jiffies > entry->expiration_date )
	     {
		  lc_nfs_printk ( "NFS cache_lookup: entry found has expired.\n" );
		  *prevptr = entry->next;
		  entry->next = nfs_lookup_cache[0].next;
		  nfs_lookup_cache[0].next = index;
		  entry->dev = 0;
		  return 0;
	     }
	     *fhandle = entry->fhandle;
	     *fattr = entry->fattr;
	     return 1;
	}

	lc_nfs_printk ( "NFS cache_lookup: found nothing\n" );
	return 0;
#endif
}

#ifdef NEWCACHE
static void nfs_lookup_cache_scavenge ( void )
{
     struct nfs_lookup_cache_entry *entry;
     int i, j, k;
     int *prev;
     int count = 0;
     int steps = 0;

     lc_nfs_printk ( "NFS:  about to scavenge\n" );

#ifdef DEBUG_NEWCACHE
     nfs_lookup_cache_check();
#endif

     lc_nfs_printk ( "NFS:  about to scavenge\n" );
     lc_nfs_printk ( "NFS:  about to scavenge\n" );

     for ( i = 0; i < NFS_LOOKUP_CACHE_BUCKETS; ++i )
     {
	  lc_nfs_printk ( "NFS: bucket %d: ", i );
	  if ( nfs_lookup_cache_buckets[i] == 0 )
	  {
	       lc_nfs_printk ( "\n" );
	       continue;
	  }

	  prev = nfs_lookup_cache_buckets + i;
	  j = *prev;
	  while ( j )
	  {
	       ++steps;
	       if ( steps > NFS_LOOKUP_CACHE_SIZE * 2 )
	       {
		    lc_nfs_printk ( "NFS: cache is inconsistent, aborting scavenge.\n" );
		    lc_nfs_printk ( "NFS: you should probably rmmod nfs quite soon.\n" );
		    return;
	       }
	       
	       lc_nfs_printk ( " %d", j );
	       entry = nfs_lookup_cache + j;
	       if ( jiffies > entry->expiration_date || entry->dev == 0 )
	       {
		    /* remove j from the list. */

		    /* save j's next, and point j's prev to it. */
		    k = entry->next;
		    *prev = k;

		    /* point the j'th entry to the first item on the free
		       list. */
		    entry->next = nfs_lookup_cache[0].next;
		    entry->dev = 0;

		    /* point the free list head to j. */
		    nfs_lookup_cache[0].next = j;

		    /* advance j. */
		    j = k;
		    
		    ++count;
	       }
	       else
	       {
		    prev = &(nfs_lookup_cache[j].next);
		    j = entry->next;
	       }
	  }
	  lc_nfs_printk ( "\n" );
     }

     lc_nfs_printk ( "NFS: scavenged %d cache entries.\n", count );

#ifdef DEBUG_NEWCACHE
     nfs_lookup_cache_check();
#endif
}
#endif

/* add a (dir,filename) entry to the cache. */

static void nfs_lookup_cache_add(struct inode *dir, const char *filename,
				 struct nfs_fh *fhandle,
				 struct nfs_fattr *fattr)
{
#ifndef NEWCACHE
	static int nfs_lookup_cache_pos = 0;
	struct nfs_lookup_cache_entry *entry;

	dfprintk(LOOKUPCACHE, "NFS: lookup_cache_add(%x/%ld, %s\n",
				dir->i_dev, dir->i_ino, filename);

        /* compensate for bug in SGI NFS server */
	if (fattr->size == -1 || fattr->uid == -1 || fattr->gid == -1
	    || fattr->atime.seconds == -1 || fattr->mtime.seconds == -1)
		return;
	if (!(entry = nfs_lookup_cache_index(dir, filename))) {
		entry = nfs_lookup_cache + nfs_lookup_cache_pos++;
		if (nfs_lookup_cache_pos == NFS_LOOKUP_CACHE_SIZE)
			nfs_lookup_cache_pos = 0;
	}

	entry->dev = dir->i_dev;
	entry->inode = dir->i_ino;
	strcpy(entry->filename, filename);
	entry->fhandle = *fhandle;
	entry->fattr = *fattr;
	entry->expiration_date = jiffies + (S_ISDIR(fattr->mode)
		? NFS_SERVER(dir)->acdirmin : NFS_SERVER(dir)->acregmin);
#else
	struct nfs_lookup_cache_entry *entry;
	int index, *prevptr;
	int myindex;
	static int tryscavengeagain = 0;
	int reindex = 0;

	if ( strlen ( filename ) > NFSLOOKUPCACHENAMELEN )
	     return;

#ifdef DEBUG_NEWCACHE
	if ( strcmp ( filename, "dump" ) == 0 )
	     return;
#endif
	
	/* compensate for bug in SGI NFS server */
	if (fattr->size == -1 || fattr->uid == -1 || fattr->gid == -1
	    || fattr->atime.seconds == -1 || fattr->mtime.seconds == -1)
		return;
	
	if ( ! ( entry = nfs_lookup_cache_index ( dir, filename,
					      &index, &prevptr ) ) )
	{
	getfree:
	     myindex = nfs_lookup_cache[0].next;
	     if ( myindex == 0 )
	     {
		  if ( tryscavengeagain > jiffies )
		  {
		       lc_nfs_printk ( "NFS: tried scavenging recently.\n" );
		       lc_nfs_printk ( "NFS: cache is REALLY full: (now: %d next: %d)\n",
				jiffies, tryscavengeagain );
		       return;
		  }
		  lc_nfs_printk ( "NFS: lookup cache is full: attempting scavenge\n" );
		  nfs_lookup_cache_scavenge();
		  reindex = 1;
		  if ( nfs_lookup_cache[0].next == 0 )
		       tryscavengeagain = jiffies + NFS_SERVER(dir)->acregmin;
		  goto getfree;
	     }
	     nfs_lookup_cache[0].next = nfs_lookup_cache[myindex].next;
	     nfs_lookup_cache[myindex].next = 0;
	     entry = nfs_lookup_cache+myindex;

	     /* scavenging may have freed the parent of the new entry,
		so reindex. */
	     
	     if ( reindex )
	     {
		  lc_nfs_printk ( "NFS: reindexing\n" );
		  nfs_lookup_cache_index ( dir, filename, &index, &prevptr );
	     }
	     
	     if ( prevptr )
	     {
		  /* index will have the last entry in the list. */
		  nfs_lookup_cache[index].next = myindex;
		  lc_nfs_printk ( "NFS cache_add: adding to chain: spot %d parent spot %d\n",
			       myindex, index );
	     }
	     else
	     {
		  /* index will be the bucket number. */
		  nfs_lookup_cache_buckets[index] = myindex;
		  lc_nfs_printk ( "NFS cache_add: starting chain: spot %d bucket %d\n",
			       myindex, index );
	     }
	}     

	entry->dev = dir->i_dev;
	entry->inode = dir->i_ino;
	strcpy(entry->filename, filename);
	entry->fhandle = *fhandle;
	entry->fattr = *fattr;
	entry->expiration_date = jiffies + (S_ISDIR(fattr->mode)
		? NFS_SERVER(dir)->acdirmin : NFS_SERVER(dir)->acregmin);

#ifdef DEBUG_NEWCACHE
	nfs_lookup_cache_check();
#endif
#endif
}

#ifdef NEWCACHE
static int nfs_lookup_cache_index_inode ( struct inode *inode,
					  int **prevptr )
{
     int i, j;
     struct nfs_lookup_cache_entry *entry;

     nfs_lookup_cache_check();
     
     for ( i = 0; i < NFS_LOOKUP_CACHE_BUCKETS; ++i )
     {
	  if ( nfs_lookup_cache_buckets[i] == 0 )
	       continue;

	  j = nfs_lookup_cache_buckets[i];
	  if ( prevptr )
	       *prevptr = nfs_lookup_cache_buckets + i;
	  while ( j )
	  {
	       entry = nfs_lookup_cache + j;
	       if ( entry->dev == inode->i_dev &&
		    entry->fattr.fileid == inode->i_ino )
	       {
		    return j;
	       }
	       j = entry->next;
	       if ( prevptr )
		    *prevptr = &(entry->next);
	  }
     } 
     if ( prevptr )
	  *prevptr = NULL;
     return 0;
}
#endif

/* remove a cache entry:  either (dir,filename) or by file inode. */

static void nfs_lookup_cache_remove(struct inode *dir, struct inode *inode,
				    const char *filename)
{
#ifndef NEWCACHE
	struct nfs_lookup_cache_entry *entry;
	kdev_t	dev;
	ino_t	fileid;
	int	i;

 	PROFILE_INC_COUNT(i_lookup_cache_remove);
 	
	if (inode) {
		dev = inode->i_dev;
		fileid = inode->i_ino;
	}
	else if ((entry = nfs_lookup_cache_index(dir, filename))) {
		dev = entry->dev;
		fileid = entry->fattr.fileid;
	}
	else
		return;

	dfprintk(LOOKUPCACHE, "NFS: lookup_cache_remove(%x/%ld)\n",
				dev, fileid);

	for (i = 0; i < NFS_LOOKUP_CACHE_SIZE; i++) {
		entry = nfs_lookup_cache + i;
 		if (entry->dev == dev && entry->fattr.fileid == fileid) {
		/* We get multiple hits in the presence of hard links,
			 and we want to remove all the items */
			entry->dev = 0;
 		}
	}
#else
	int index, *prevptr;
	struct nfs_lookup_cache_entry *entry;

	lc_nfs_printk ( "NFS: lookup_cache_remove()\n" );
	
	PROFILE_INC_COUNT(i_lookup_cache_remove);

	if ( inode )
	{
	     index = nfs_lookup_cache_index_inode ( inode, &prevptr );
	}
	else
	{
	     entry = nfs_lookup_cache_index ( dir, filename, &index,
					      &prevptr );
	     if ( entry == NULL )
		  index = 0;
	}

	if ( index )
	{
	     entry = nfs_lookup_cache + index;
	     
	     *prevptr = entry->next;
	     entry->next = nfs_lookup_cache[0].next;
	     nfs_lookup_cache[0].next = index;
	}
#endif
}

void
nfs_refresh_inode(struct inode *inode, struct nfs_fattr *fattr)
{
	lc_nfs_printk ( "NFS: call to nfs_refresh_inode\n" );
	if (!inode || !fattr) {
	  printk("nfs_refresh_inode: inode or fattr is NULL\n");
	  return;
	}
	if (inode->i_ino != fattr->fileid) {
	  printk("nfs_refresh_inode: inode number mismatch\n");
	  return;
	}
	nfs_refresh_inode_not_cache(inode,fattr);
	nfs_lookup_cache_refresh(inode, fattr);
}

/* search for a file in the cache, by inode, update its attributes. */

void nfs_lookup_cache_refresh(struct inode *file,
			      struct nfs_fattr *fattr)
{
#ifndef NEWCACHE
	struct nfs_lookup_cache_entry *entry;
	kdev_t dev = file->i_dev;
	int fileid = file->i_ino;
	int i;

 	PROFILE_INC_COUNT(i_lookup_cache_refresh);
 	
	for (i = 0; i < NFS_LOOKUP_CACHE_SIZE; i++) {
		entry = nfs_lookup_cache + i;
		if (entry->dev == dev && entry->fattr.fileid == fileid) {
		/* We get multiple hits in the presence of hard links,
			 and we want to update all the attributes */
			entry->fattr = *fattr;
		}
	}
#else
	int index;

	PROFILE_INC_COUNT(i_lookup_cache_refresh);

	/* Note-- nfs_lookup_cache_index_inode is expensive -- it does
	   a linear scan of the whole hash */
	lc_nfs_printk ( "NFS: REFRESHING BY INODE (%p).\n", file );
	index = nfs_lookup_cache_index_inode ( file, NULL );
	lc_nfs_printk ( "NFS:   refresh index is %d (%s)\n", index, index==0?"":nfs_lookup_cache[index].filename );
	if ( index )
	     nfs_lookup_cache[index].fattr = *fattr;
	lc_nfs_printk ( "NFS: INODE REFRESH DONE\nNFS:\n" );
#endif
}

/* just update the lookup cache, given the name
   this is done because it's much less efficient when
   all we've got is the inode */
void
nfs_refresh_inode_cache_name(struct inode *inodeDir, const char *name,struct nfs_fattr *fattr)
{
#ifdef NEWCACHE
	int index, *prevptr;
	struct nfs_lookup_cache_entry *entry = NULL;
	lc_nfs_printk ( "NFS: REFRESHING BY NAME (%p,%s)\n", inodeDir, name );
	entry = nfs_lookup_cache_index(inodeDir,name,&index,&prevptr);
	lc_nfs_printk ( "NFS:   refreshing entry %d (%s)\n", (entry==NULL)?-1:index,
		 (entry==NULL)?"":entry->filename );
#else
	struct nfs_lookup_cache_entry *entry = nfs_lookup_cache_index(inodeDir,name);
#endif
	if ( entry )
	     entry->fattr = *fattr;
	lc_nfs_printk ( "NFS: NAME REFRESH DONE.\nNFS:\n" );
}

#if defined(NFS_INSTRUMENT_ALL_LOOKUPS) || defined(NFS_INSTRUMENT_RPC_LOOKUPS)
static void jiffiestosecms(unsigned long jiffies, unsigned long *psec, int *pms)
{
	unsigned long jif = jiffies;
	*pms = (jif % HZ) * (100.0 / HZ);
	*psec = jif / HZ;
	return;
}
#endif

static inline char
ch_from_nfstype(int i)
{
	switch(i) {
	case NFREG:
		return '.';
	case NFDIR:
		return '/';
	case NFFIFO:
		return '|';
	case NFLNK:
		return '@';
	case NFBLK:
		return 'b';
	case NFCHR:
		return 'c';
	case NFSOCK:
		return '#';
	default:
		break;
	}
	return '?';
}


static int nfs_lookup(struct inode *dir, const char *__name, int len,
		      struct inode **result)
{
	struct nfs_fh fhandle;
	struct nfs_fattr fattr;
	char name[len > NFS_MAXNAMLEN? 1 : len+1];
	int error;
#if defined(NFS_INSTRUMENT_ALL_LOOKUPS) || defined(NFS_INSTRUMENT_RPC_LOOKUPS)
	unsigned long sec;
	int ms;
#endif

 	PROFILE_INC_COUNT(i_lookup);

#if defined(NFS_INSTRUMENT_ALL_LOOKUPS) || defined(NFS_INSTRUMENT_RPC_LOOKUPS)
	jiffiestosecms(jiffies,&sec,&ms);
#endif
#ifdef NFS_INSTRUMENT_ALL_LOOKUPS
	printk("NFS: lookup(%x/%ld, %.*s) @ %ld.%d\n",
				dir->i_dev, dir->i_ino, len, __name, sec, ms);
#else
	dfprintk(VFS,"NFS: lookup(%x/%ld, %.*s)\n",
					 dir->i_dev, dir->i_ino, len, __name);
#endif

	*result = NULL;
	if (!dir || !S_ISDIR(dir->i_mode)) {
		printk("nfs_lookup: inode is NULL or not a directory\n");
		iput(dir);
		return -ENOENT;
	}
	if (len > NFS_MAXNAMLEN) {
		iput(dir);
		return -ENAMETOOLONG;
	}
	memcpy(name,__name,len);
	name[len] = '\0';
	if (len == 1 && name[0] == '.') { /* cheat for "." */
		*result = dir;
		return 0;
	}
	if ((NFS_SERVER(dir)->flags & NFS_MOUNT_NOAC)
	    || !nfs_lookup_cache_lookup(dir, name, &fhandle, &fattr)) {
		PROFILE_INC_COUNT(i_lookup_miss);
		if (nfs_async_lookups_enabled &&
				(error = nfs_failed_lookup_cache_lookup(dir,name))) {
			iput(dir);
			return error;
		}
		error = nfs_proc_lookup(NFS_SERVER(dir), dir->i_sb, NFS_FH(dir),
														name, &fhandle, &fattr);
#ifdef NFS_INSTRUMENT_RPC_LOOKUPS
		printk("NFS: lookup(%x/%ld, %.*s) %ld.%02d, type %c, returned %d\n",
					 dir->i_dev, dir->i_ino, len, __name, sec, ms,
					 ch_from_nfstype(fattr.type),
					 error);
#endif
		if (error) {
#ifdef ASYNCH_PREDICT_FAILED_LOOKUPS
			if (error == -ENOENT && nfs_async_lookups_enabled)
				nfs_handle_failed_lookup(dir,name,error);
#endif
			iput(dir);
			return error;
		}
		nfs_lookup_cache_add(dir, name, &fhandle, &fattr);
		NFS_READTIME(inode_from_sb_fileid(dir->i_sb,fattr.fileid)) = jiffies;
	}
	if (!(*result = nfs_fhget(dir->i_sb, &fhandle, &fattr))) {
		iput(dir);
		return -EACCES;
	}
	iput(dir);
	return 0;
}

static int nfs_create(struct inode *dir, const char *name, int len, int mode,
		      struct inode **result)
{
	struct nfs_sattr sattr;
	struct nfs_fattr fattr;
	struct nfs_fh fhandle;
	int error;

	dfprintk(VFS, "NFS: create(%x/%ld, %s\n",
				dir->i_dev, dir->i_ino, name);
 	PROFILE_INC_COUNT(i_create);

	*result = NULL;
	if (!dir || !S_ISDIR(dir->i_mode)) {
		printk("nfs_create: inode is NULL or not a directory\n");
		iput(dir);
		return -ENOENT;
	}
	if (len > NFS_MAXNAMLEN) {
		iput(dir);
		return -ENAMETOOLONG;
	}
	sattr.mode = mode;
	sattr.uid = sattr.gid = sattr.size = (unsigned) -1;
	sattr.atime.seconds = sattr.mtime.seconds = (unsigned) -1;
	if ((error = nfs_proc_create(NFS_SERVER(dir), NFS_FH(dir),
		name, &sattr, &fhandle, &fattr))) {
		iput(dir);
		return error;
	}
	if (!(*result = nfs_fhget(dir->i_sb, &fhandle, &fattr))) {
		iput(dir);
		return -EACCES;
	}
	nfs_lookup_cache_add(dir, name, &fhandle, &fattr);
	nfs_invalidate_dircache(dir);
	iput(dir);
	return 0;
}

static int nfs_mknod(struct inode *dir, const char *name, int len,
		     int mode, int rdev)
{
	struct nfs_sattr sattr;
	struct nfs_fattr fattr;
	struct nfs_fh fhandle;
	int error;

	dfprintk(VFS, "NFS: mknod(%x/%ld, %s\n",
				dir->i_dev, dir->i_ino, name);
 	PROFILE_INC_COUNT(i_mknod);

	if (!dir || !S_ISDIR(dir->i_mode)) {
		printk("nfs_mknod: inode is NULL or not a directory\n");
		iput(dir);
		return -ENOENT;
	}
	if (len > NFS_MAXNAMLEN) {
		iput(dir);
		return -ENAMETOOLONG;
	}
	sattr.mode = mode;
	sattr.uid = sattr.gid = (unsigned) -1;
	if (S_ISCHR(mode) || S_ISBLK(mode))
		sattr.size = rdev; /* get out your barf bag */
	else
		sattr.size = (unsigned) -1;
	sattr.atime.seconds = sattr.mtime.seconds = (unsigned) -1;
	error = nfs_proc_create(NFS_SERVER(dir), NFS_FH(dir),
		name, &sattr, &fhandle, &fattr);
	if (!error)
		nfs_lookup_cache_add(dir, name, &fhandle, &fattr);
	nfs_invalidate_dircache(dir);
	iput(dir);
	return error;
}

static int nfs_mkdir(struct inode *dir, const char *name, int len, int mode)
{
	struct nfs_sattr sattr;
	struct nfs_fattr fattr;
	struct nfs_fh fhandle;
	int error;

	dfprintk(VFS, "NFS: mkdir(%x/%ld, %s\n",
				dir->i_dev, dir->i_ino, name);
	PROFILE_INC_COUNT(i_mkdir);

	if (!dir || !S_ISDIR(dir->i_mode)) {
		printk("nfs_mkdir: inode is NULL or not a directory\n");
		iput(dir);
		return -ENOENT;
	}
	if (len > NFS_MAXNAMLEN) {
		iput(dir);
		return -ENAMETOOLONG;
	}
	sattr.mode = mode;
	sattr.uid = sattr.gid = sattr.size = (unsigned) -1;
	sattr.atime.seconds = sattr.mtime.seconds = (unsigned) -1;
	error = nfs_proc_mkdir(NFS_SERVER(dir), NFS_FH(dir),
		name, &sattr, &fhandle, &fattr);
	if (!error)
		nfs_lookup_cache_add(dir, name, &fhandle, &fattr);
	nfs_invalidate_dircache(dir);
	iput(dir);
	return error;
}

static int nfs_rmdir(struct inode *dir, const char *name, int len)
{
	int error;

	dfprintk(VFS, "NFS: rmdir(%x/%ld, %s\n",
				dir->i_dev, dir->i_ino, name);
	PROFILE_INC_COUNT(i_rmdir);

	if (!dir || !S_ISDIR(dir->i_mode)) {
		printk("nfs_rmdir: inode is NULL or not a directory\n");
		iput(dir);
		return -ENOENT;
	}
	if (len > NFS_MAXNAMLEN) {
		iput(dir);
		return -ENAMETOOLONG;
	}
	error = nfs_proc_rmdir(NFS_SERVER(dir), NFS_FH(dir), name);
	if (!error)
		nfs_lookup_cache_remove(dir, NULL, name);
	nfs_invalidate_dircache(dir);
	iput(dir);
	return error;
}

static int nfs_sillyrename(struct inode *dir, const char *name, int len)
{
	struct inode	*inode;
	char		silly[16];
	int		slen, ret;

	dir->i_count++;
	if (nfs_lookup(dir, name, len, &inode) < 0)
		return -EIO;		/* arbitrary */

	if (inode->i_count == 1) {
		iput(inode);
		return -EIO;
	}
	if (NFS_RENAMED_DIR(inode)) {
		iput(NFS_RENAMED_DIR(inode));
		NFS_RENAMED_DIR(inode) = NULL;
		iput(inode);
		return -EIO;
	}

	slen = sprintf(silly, ".nfs%ld", inode->i_ino);
	if (len == slen && !strncmp(name, silly, len)) {
		iput(inode);
		return -EIO;		/* DWIM */
	}

	dfprintk(VFS, "NFS: sillyrename(%x/%ld, %s)\n",
				dir->i_dev, dir->i_ino, name);

	ret = nfs_proc_rename(NFS_SERVER(dir), NFS_FH(dir), name,
					       NFS_FH(dir), silly);
	if (ret >= 0) {
		nfs_lookup_cache_remove(dir, NULL, name);
		nfs_lookup_cache_remove(dir, NULL, silly);
		NFS_RENAMED_DIR(inode) = dir;
		dir->i_count++;
	}
	nfs_invalidate_dircache(dir);
	iput(inode);
	return ret;
}

/*
 * When releasing the inode, finally remove any unlinked but open files.
 * Note that we have to clear the set of pending signals temporarily;
 * otherwise the RPC call will fail.
 */
void nfs_sillyrename_cleanup(struct inode *inode)
{
	unsigned long	oldsig;
	struct inode	*dir = NFS_RENAMED_DIR(inode);
	char		silly[14];
	int		error, slen;

	dfprintk(VFS, "NFS: sillyrename cleanup(%x/%ld)\n",
				inode->i_dev, inode->i_ino);

	oldsig = current->signal;
	current->signal = 0;

	slen = sprintf(silly, ".nfs%ld", inode->i_ino);
	error = nfs_proc_remove(NFS_SERVER(dir), NFS_FH(dir), silly);
	if (error < 0)
		printk("NFS: silly_rename cleanup failed (err %d)\n", -error);

	nfs_lookup_cache_remove(dir, NULL, silly);
	nfs_invalidate_dircache(dir);
	NFS_RENAMED_DIR(inode) = NULL;
	iput(dir);

	current->signal |= oldsig;
}

static int nfs_unlink(struct inode *dir, const char *name, int len)
{
	int error;

	dfprintk(VFS, "NFS: unlink(%x/%ld, %s)\n",
				dir->i_dev, dir->i_ino, name);
	PROFILE_INC_COUNT(i_unlink);

	if (!dir || !S_ISDIR(dir->i_mode)) {
		printk("nfs_unlink: inode is NULL or not a directory\n");
		iput(dir);
		return -ENOENT;
	}
	if (len > NFS_MAXNAMLEN) {
		iput(dir);
		return -ENAMETOOLONG;
	}
	if ((error = nfs_sillyrename(dir, name, len)) < 0) {
		error = nfs_proc_remove(NFS_SERVER(dir), NFS_FH(dir), name);
		if (!error)
			nfs_lookup_cache_remove(dir, NULL, name);
	}
	nfs_invalidate_dircache(dir);
	iput(dir);
	return error;
}

static int nfs_symlink(struct inode *dir, const char *name, int len,
		       const char *symname)
{
	struct nfs_sattr sattr;
	int error;

	dfprintk(VFS, "NFS: symlink(%x/%ld, %s, %s)\n",
				dir->i_dev, dir->i_ino, name, symname);
	PROFILE_INC_COUNT(i_symlink);

	if (!dir || !S_ISDIR(dir->i_mode)) {
		printk("nfs_symlink: inode is NULL or not a directory\n");
		iput(dir);
		return -ENOENT;
	}
	if (len > NFS_MAXNAMLEN) {
		iput(dir);
		return -ENAMETOOLONG;
	}
	if (strlen(symname) > NFS_MAXPATHLEN) {
		iput(dir);
		return -ENAMETOOLONG;
	}
	sattr.mode = S_IFLNK | S_IRWXUGO; /* SunOS 4.1.2 crashes without this! */
	sattr.uid = sattr.gid = sattr.size = (unsigned) -1;
	sattr.atime.seconds = sattr.mtime.seconds = (unsigned) -1;
	error = nfs_proc_symlink(NFS_SERVER(dir), NFS_FH(dir),
		name, symname, &sattr);
	nfs_invalidate_dircache(dir);
	iput(dir);
	return error;
}

static int nfs_link(struct inode *oldinode, struct inode *dir,
		    const char *name, int len)
{
	int error;

	dfprintk(VFS, "NFS: link(%x/%ld -> %x/%ld, %s)\n",
				oldinode->i_dev, oldinode->i_ino,
				dir->i_dev, dir->i_ino, name);
	PROFILE_INC_COUNT(i_link);

	if (!oldinode) {
		printk("nfs_link: old inode is NULL\n");
		iput(oldinode);
		iput(dir);
		return -ENOENT;
	}
	if (!dir || !S_ISDIR(dir->i_mode)) {
		printk("nfs_link: dir is NULL or not a directory\n");
		iput(oldinode);
		iput(dir);
		return -ENOENT;
	}
	if (len > NFS_MAXNAMLEN) {
		iput(oldinode);
		iput(dir);
		return -ENAMETOOLONG;
	}
	error = nfs_proc_link(NFS_SERVER(oldinode), NFS_FH(oldinode),
		NFS_FH(dir), name);
	if (!error) {
		nfs_lookup_cache_remove(dir, oldinode, NULL);
		NFS_READTIME(oldinode) = 0;	/* force getattr */
	}
	nfs_invalidate_dircache(dir);
	iput(oldinode);
	iput(dir);
	return error;
}

/*
 * RENAME
 * FIXME: Some nfsds, like the Linux user space nfsd, may generate a
 * different file handle for the same inode after a rename (e.g. when
 * moving to a different directory). A fail-safe method to do so would
 * be to look up old_dir/old_name, create a link to new_dir/new_name and
 * rename the old file using the silly_rename stuff. This way, the original
 * file in old_dir will go away when the last process iput()s the inode.
 */
static int nfs_rename(struct inode *old_dir, const char *old_name, int old_len,
		      struct inode *new_dir, const char *new_name, int new_len,
		      int must_be_dir)
{
	int error;

	dfprintk(VFS, "NFS: rename(%x/%ld, %s -> %x/%ld, %s)\n",
				old_dir->i_dev, old_dir->i_ino, old_name,
				new_dir->i_dev, new_dir->i_ino, new_name);
	PROFILE_INC_COUNT(i_rename);

	if (!old_dir || !S_ISDIR(old_dir->i_mode)) {
		printk("nfs_rename: old inode is NULL or not a directory\n");
		iput(old_dir);
		iput(new_dir);
		return -ENOENT;
	}
	if (!new_dir || !S_ISDIR(new_dir->i_mode)) {
		printk("nfs_rename: new inode is NULL or not a directory\n");
		iput(old_dir);
		iput(new_dir);
		return -ENOENT;
	}
	if (old_len > NFS_MAXNAMLEN || new_len > NFS_MAXNAMLEN) {
		iput(old_dir);
		iput(new_dir);
		return -ENAMETOOLONG;
	}

	/* We don't do rename() with trailing slashes over NFS now. Hmm. */
	if (must_be_dir)
		return -EINVAL;

	error = nfs_proc_rename(NFS_SERVER(old_dir),
		NFS_FH(old_dir), old_name,
		NFS_FH(new_dir), new_name);
	if (!error) {
		nfs_lookup_cache_remove(old_dir, NULL, old_name);
		nfs_lookup_cache_remove(new_dir, NULL, new_name);
	}
	nfs_invalidate_dircache(old_dir);
	nfs_invalidate_dircache(new_dir);
	iput(old_dir);
	iput(new_dir);
	return error;
}

/*
 * Many nfs protocol calls return the new file attributes after
 * an operation.  Here we update the inode to reflect the state
 * of the server's inode.
 */

void nfs_refresh_inode_not_cache(struct inode *inode, struct nfs_fattr *fattr)
{
	int was_empty;
	BOOL fInvalidatePages = FALSE;

	dfprintk(VFS, "NFS: refresh_inode(%x/%ld ct=%d)\n",
				inode->i_dev, inode->i_ino, inode->i_count);

	was_empty = (inode->i_mode == 0);
	inode->i_mode = fattr->mode;
	inode->i_nlink = fattr->nlink;
	inode->i_uid = fattr->uid;
	inode->i_gid = fattr->gid;

	/* Size changed from outside: invalidate caches on next read */
	/* FIXGJB: why? This seems silly, since we know the real size --
		 we  want to be sure we do not use the old pages, but
		 needn't redo the getattr */
	if (inode->i_size != fattr->size) {
		dfprintk(PAGECACHE, "NFS:      cacheinv(%x/%ld)\n",
					inode->i_dev, inode->i_ino);
		fInvalidatePages = TRUE;
#ifdef FIXGJB_WHY
		NFS_CACHEINV(inode);
#endif
	}
	if (NFS_OLDMTIME(inode) != fattr->mtime.seconds) {
		dfprintk(PAGECACHE, "NFS:      mtime change on %x/%ld\n",
					inode->i_dev, inode->i_ino);
		NFS_ATTRTIMEO(inode) = NFS_MINATTRTIMEO(inode);
		fInvalidatePages = TRUE;
	}
	inode->i_size = fattr->size;
	if (S_ISCHR(inode->i_mode) || S_ISBLK(inode->i_mode))
		inode->i_rdev = to_kdev_t(fattr->rdev);
	else
		inode->i_rdev = 0;
	inode->i_blocks = fattr->blocks;
	inode->i_atime = fattr->atime.seconds;
	inode->i_mtime = fattr->mtime.seconds;
	inode->i_ctime = fattr->ctime.seconds;
	if (S_ISREG(inode->i_mode))
		inode->i_op = &nfs_file_inode_operations;
	else if (S_ISDIR(inode->i_mode))
		inode->i_op = &nfs_dir_inode_operations;
	else if (S_ISLNK(inode->i_mode))
		inode->i_op = &nfs_symlink_inode_operations;
	else if (S_ISCHR(inode->i_mode))
		inode->i_op = &chrdev_inode_operations;
	else if (S_ISBLK(inode->i_mode))
		inode->i_op = &blkdev_inode_operations;
	else if (S_ISFIFO(inode->i_mode)) {
		if (was_empty)
			init_fifo(inode);
	} else
		inode->i_op = NULL;

	if (fInvalidatePages) {
		/* we need to invalidate */
		wnfs_printk("NFS: %s, invalidating pages for (%d,%ld)\n",__FUNCTION__,
								inode->i_dev,inode->i_ino);
		if (!S_ISDIR(inode->i_mode)) {
			/* This sends off all dirty pages off to the server.
			 * Note that this function must not sleep. */
			nfs_invalidate_pages(inode);
			invalidate_inode_pages(inode);
		} else {
			nfs_invalidate_dircache(inode);
		}

		/* reset timeouts */
		NFS_OLDMTIME(inode)  = fattr->mtime.seconds;
		NFS_ATTRTIMEO(inode) = NFS_MINATTRTIMEO(inode);
	} else {
		/* did not invalidate, so Update attrtimeo value */
		if ((NFS_ATTRTIMEO(inode) <<= 1) > NFS_MAXATTRTIMEO(inode))
			NFS_ATTRTIMEO(inode) = NFS_MAXATTRTIMEO(inode);
	}
	NFS_READTIME(inode) = jiffies;
}

/* Local Variables: */
/* compile-command: "cd ../..; make modules" */
/* tab-width: 8 */
/* End: */
