/*
 *  linux/fs/nfs/inode.c
 *
 *  Copyright (C) 1992  Rick Sladkey
 *
 *  nfs inode and superblock handling functions
 *
 *  Modularised by Alan Cox <Alan.Cox@linux.org>, while hacking some
 *  experimental NFS changes. Modularisation taken straight from SYS5 fs.
 *
 *  Change to nfs_read_super() to permit NFS mounts to multi-homed hosts.
 *  J.S.Peatfield@damtp.cam.ac.uk
 *
 * Misc. extension (including clear_inode_hook) added by 
 * Greg J. Badros <gjb@cs.washington.edu> to support local disk
 * caching
 */

#include <linux/config.h>
#include <linux/module.h>

#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/errno.h>
#include <linux/locks.h>
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>
#include <linux/sunrpc/clnt.h>
#include <linux/nfs_fs.h>
#include <linux/lockd/bind.h>

#include <asm/system.h>
#ifdef KERNEL_2_1
#include <asm/uaccess.h>
#else
#define copy_to_user(to, from, size) memcpy_tofs(to, from, size)
#define lockd_up() ;
#define lockd_down() ;
#endif

#include <linux/proc_fs.h>
#include <linux/file.h>
#include "dir.h"
#include "profile.h"
#include "nfs_cache.h"
#include "nfs_fillind.h"

#define NFSDBG_FACILITY		NFSDBG_VFS
#define NFS_USE_FILLIND

static int nfs_notify_change(struct inode *, struct iattr *);
static void nfs_put_inode(struct inode *);
static void nfs_put_super(struct super_block *);
static void nfs_read_inode(struct inode *);
static void nfs_statfs(struct super_block *, struct statfs *, int bufsiz);

static struct super_operations nfs_sops = { 
	nfs_read_inode,		/* read inode */
	nfs_notify_change,	/* notify change */
	NULL,			/* write inode */
	nfs_put_inode,		/* put inode */
	nfs_put_super,		/* put superblock */
	NULL,			/* write superblock */
	nfs_statfs,		/* stat filesystem */
	NULL
};

/*
 * The "read_inode" function doesn't actually do anything:
 * the real data is filled in later in nfs_fhget. Here we
 * just mark the cache times invalid, and zero out i_mode
 * (the latter makes "nfs_refresh_inode" do the right thing
 * wrt pipe inodes)
 */
static void
nfs_read_inode(struct inode * inode)
{
	dprintk("NFS: read_inode(%x/%ld)\n", inode->i_dev, inode->i_ino);
	PROFILE_INC_COUNT(i_read_inode);

	inode->i_blksize = inode->i_sb->s_blocksize;
	inode->i_mode = 0;
	inode->i_op = NULL;
	inode->clear_inode_hook = NULL; /* we point it's clear_inode_hook
																		 when we open the cache file */
	NFS_CACHE_INODE(inode) = NULL;
	NFS_CACHEINV(inode);
}


static void
nfs_put_inode(struct inode * inode)
{
	dprintk("NFS: put_inode(%x/%ld)\n", inode->i_dev, inode->i_ino);
	PROFILE_INC_COUNT(i_put_inode);

	if (NFS_RENAMED_DIR(inode))
		nfs_sillyrename_cleanup(inode);
	if (inode->i_pipe)
		clear_inode(inode);
}

void
nfs_put_super(struct super_block *sb)
{

	struct rpc_clnt	*rpc;
	nfs_printk("NFS: nfs_put_super called\n");
	PROFILE_INC_COUNT(i_put_super);

	if ((rpc = sb->u.nfs_sb.s_server.client) != NULL)
		rpc_shutdown_client(rpc);

	invalidate_inodes(sb->s_dev);
	/* clear_inodeCache_hook takes care of the inodeNFSs, but
	   we still need to clean up the inodeCache-s that are partial */
	nfs_cleanup_inodes_with_partial_caches(sb->s_dev, FALSE /* no unlink files */);

	if (sb->u.nfs_sb.s_cache != NULL) {
		nfs_unuse_cache(sb);
	}

	lockd_down();		/* release rpc.lockd */
	rpciod_down();		/* release rpciod */
	lock_super(sb);
	sb->s_dev = 0;
	unlock_super(sb);
	MOD_DEC_USE_COUNT;
}

/*
 * Compute and set NFS server blocksize
 */
static unsigned int
nfs_block_size(unsigned int bsize, unsigned char *nrbitsp)
{
	if (bsize < 1024)
		bsize = NFS_DEF_FILE_IO_BUFFER_SIZE;
	else if (bsize >= NFS_MAX_FILE_IO_BUFFER_SIZE)
		bsize = NFS_MAX_FILE_IO_BUFFER_SIZE;

	/* make sure blocksize is a power of two */
	if ((bsize & (bsize - 1)) || nrbitsp) {
		unsigned int	nrbits;

		for (nrbits = 31; nrbits && !(bsize & (1 << nrbits)); nrbits--)
			;
		bsize = 1 << nrbits;
		if (nrbitsp)
			*nrbitsp = nrbits;
		if (bsize < NFS_DEF_FILE_IO_BUFFER_SIZE)
			bsize = NFS_DEF_FILE_IO_BUFFER_SIZE;
	}

	return bsize;
}

/*
 * The way this works is that the mount process passes a structure
 * in the data argument which contains the server's IP address
 * and the root file handle obtained from the server's mount
 * daemon. We stash these away in the private superblock fields.
 */
struct super_block *
nfs_read_super(struct super_block *sb, void *raw_data, int silent)
{
	struct nfs_mount_data	*data = (struct nfs_mount_data *) raw_data;
	struct sockaddr_in	srvaddr;
	struct nfs_server	*server;
	struct rpc_timeout	timeparms;
	struct rpc_xprt		*xprt;
	struct rpc_clnt		*clnt;
	unsigned int		authflavor;
	int			tcp;
	kdev_t			dev = sb->s_dev;

	verb_nfs_printk("NFS: nfs_read_super called\n");
	MOD_INC_USE_COUNT;
	if (!data) {
		printk("nfs_read_super: missing data argument\n");
		sb->s_dev = 0;
		MOD_DEC_USE_COUNT;
		return NULL;
	}
	if (data->version != NFS_MOUNT_VERSION) {
		printk("nfs warning: mount version %s than kernel\n",
			data->version < NFS_MOUNT_VERSION ? "older" : "newer");
		if (data->version < 2)
			data->namlen = 0;
		if (data->version < 3)
			data->bsize  = 0;
	}

	lock_super(sb);

	server           = &sb->u.nfs_sb.s_server;
	sb->s_magic      = NFS_SUPER_MAGIC;
	sb->s_dev        = dev;
	sb->s_op         = &nfs_sops;
	sb->s_blocksize  = nfs_block_size(data->bsize, &sb->s_blocksize_bits);
	server->rsize    = nfs_block_size(data->rsize, NULL);
	server->wsize    = nfs_block_size(data->wsize, NULL);
	server->flags    = data->flags;
	server->acregmin = data->acregmin*HZ;
	server->acregmax = data->acregmax*HZ;
	server->acdirmin = data->acdirmin*HZ;
	server->acdirmax = data->acdirmax*HZ;
	strcpy(server->hostname, data->hostname);
	sb->u.nfs_sb.s_root = data->root;

	sb->u.nfs_sb.s_cache = NULL;

	/* now we need to get the inode for the caching directory --gjb */
	if (server->flags & NFS_MOUNT_DC_ENABLE) {
		verb_nfs_printk("NFS: nfs_read_super, caching is on, trying to get dir, etc.\n");
		if (nfs_set_cache_dir_and_fullstate(server,data,&sb->u.nfs_sb) == 0) {
			/* success (failures have printk-s in nfs_set_cache_dir_and_fullstate fn */
			nfs_printk("NFS: nfs_read_super, using cache at %lx\n\
NFS:... root = %s, min_free = %d, max_size = %d\n",
								 (long unsigned) sb->u.nfs_sb.s_cache,
								 sb->u.nfs_sb.s_cache->root,
								 sb->u.nfs_sb.s_cache->min_free,
								 sb->u.nfs_sb.dc_max_size);
			/* reset the full bit */
			sb->u.nfs_sb.s_cache->inodeFullStateIndicator->i_mode &= 
				~(NFS_MAX_SPACE_VIOL_BIT | NFS_MIN_FREE_VIOL_BIT | NFS_ERROR_WRITING_BIT);
			sb->u.nfs_sb.s_cache->inodeFullStateIndicator->i_dirt = 1;
			/* and check whether it should be set */
			nfs_check_too_full(sb);
		} 
	}
	
	/* We now require that the mount process passes the remote address */
	memcpy(&srvaddr, &data->addr, sizeof(srvaddr));
	if (srvaddr.sin_addr.s_addr == INADDR_ANY) {
		printk("NFS: mount program didn't pass remote address!\n");
		MOD_DEC_USE_COUNT;
		return NULL;
	}

	/* Which protocol do we use? */
	tcp   = (data->flags & NFS_MOUNT_TCP);

	/* Initialize timeout values */
	timeparms.to_initval = data->timeo * HZ / 10;
	timeparms.to_retries = data->retrans;
	nfs_printk("NFS: timeo=%d 10ths of sec.,retrans=%d\n",data->timeo,data->retrans);
	timeparms.to_maxval  = tcp? RPC_MAX_TCP_TIMEOUT : RPC_MAX_UDP_TIMEOUT;
	timeparms.to_exponential = 1;

	/* Choose authentication flavor */
	if (data->flags & NFS_MOUNT_SECURE) {
		authflavor = RPC_AUTH_DES;
	} else if (data->flags & NFS_MOUNT_KERBEROS) {
		authflavor = RPC_AUTH_KRB;
	} else {
		authflavor = RPC_AUTH_UNIX;
	}

	/* Now create transport and client */
	xprt = xprt_create_proto(tcp? IPPROTO_TCP : IPPROTO_UDP,
						&srvaddr, &timeparms);
	if (xprt == NULL) {
		printk("NFS: cannot create RPC transport.\n");
		goto failure;
	}

	clnt = rpc_create_client(xprt, server->hostname, &nfs_program,
						NFS_VERSION, authflavor);
	if (clnt == NULL) {
		printk("NFS: cannot create RPC client.\n");
		xprt_destroy(xprt);
		goto failure;
	}

	clnt->cl_intr     = (data->flags & NFS_MOUNT_INTR)? 1 : 0;
	clnt->cl_softrtry = (data->flags & NFS_MOUNT_SOFT)? 1 : 0;
	clnt->cl_chatty   = 1;
	server->client    = clnt;

	/* Fire up rpciod if not yet running */
	rpciod_up();

	/* Unlock super block and try to get root fh attributes */
	unlock_super(sb);

	if ((sb->s_mounted = nfs_fhget(sb, &data->root, NULL)) != NULL) {
		/* We're airborne */
		lockd_up();
		return sb;
	}

	/* Yargs. It didn't work out. */
	printk("nfs_read_super: get root inode failed\n");
	rpc_shutdown_client(server->client);
	rpciod_down();

failure:
	MOD_DEC_USE_COUNT;
	if (sb->s_lock)
		unlock_super(sb);
	sb->s_dev = 0;
	return NULL;
}

static void
nfs_statfs(struct super_block *sb, struct statfs *buf, int bufsiz)
{
	int error;
	struct nfs_fsinfo res;
	struct statfs tmp;

	PROFILE_INC_COUNT(i_statfs);

	error = nfs_proc_statfs(&sb->u.nfs_sb.s_server, &sb->u.nfs_sb.s_root,
		&res);
	if (error) {
		printk("nfs_statfs: statfs error = %d\n", -error);
		res.bsize = res.blocks = res.bfree = res.bavail = 0;
	}
	tmp.f_type = NFS_SUPER_MAGIC;
	tmp.f_bsize = res.bsize;
	tmp.f_blocks = res.blocks;
	tmp.f_bfree = res.bfree;
	tmp.f_bavail = res.bavail;
	tmp.f_files = 0;
	tmp.f_ffree = 0;
	tmp.f_namelen = NAME_MAX;
	copy_to_user(buf, &tmp, bufsiz);
}


/*
 * This is our own version of iget that looks up inodes by file handle
 * instead of inode number.  We use this technique instead of using
 * the vfs read_inode function because there is no way to pass the
 * file handle or current attributes into the read_inode function.
 * We just have to be careful not to subvert iget's special handling
 * of mount points.
 */
struct inode *
nfs_fhget(struct super_block *sb, struct nfs_fh *fhandle,
				  struct nfs_fattr *fattr)
{
	struct nfs_fattr newfattr;
	int error;
	struct inode *inode;

	if (!sb) {
		printk("nfs_fhget: super block is NULL\n");
		return NULL;
	}
	if (!fattr) {
		verb_wnfs_printk("NFS: nfs_fhget: calling getattr\n");
		error = nfs_proc_getattr(&sb->u.nfs_sb.s_server, fhandle,
			&newfattr);
		if (error) {
			printk("nfs_fhget: getattr error = %d\n", -error);
			return NULL;
		}
		fattr = &newfattr;
	}
	if (!(inode = iget(sb, fattr->fileid))) {
		printk("nfs_fhget: iget failed\n");
		return NULL;
	}
	if (inode->i_dev == sb->s_dev) {
		if (inode->i_ino != fattr->fileid) {
			printk("nfs_fhget: unexpected inode from iget\n");
			return inode;
		}
		*NFS_FH(inode) = *fhandle;
		nfs_refresh_inode(inode, fattr);
	}
	dprintk("NFS: fhget(%x/%ld ct=%d)\n",
				inode->i_dev, inode->i_ino, inode->i_count);

	return inode;
}

int
nfs_notify_change(struct inode *inode, struct iattr *attr)
{
	struct nfs_sattr sattr;
	struct nfs_fattr fattr;
	int error;

	PROFILE_INC_COUNT(i_notify_change);

	sattr.mode = (u32) -1;
	if (attr->ia_valid & ATTR_MODE) 
		sattr.mode = attr->ia_mode;

	sattr.uid = (u32) -1;
	if (attr->ia_valid & ATTR_UID)
		sattr.uid = attr->ia_uid;

	sattr.gid = (u32) -1;
	if (attr->ia_valid & ATTR_GID)
		sattr.gid = attr->ia_gid;


	sattr.size = (u32) -1;
	if ((attr->ia_valid & ATTR_SIZE) && S_ISREG(inode->i_mode))
		sattr.size = attr->ia_size;

	sattr.mtime.seconds = sattr.mtime.useconds = (u32) -1;
	if (attr->ia_valid & ATTR_MTIME) {
		sattr.mtime.seconds = attr->ia_mtime;
		sattr.mtime.useconds = 0;
	}

	sattr.atime.seconds = sattr.atime.useconds = (u32) -1;
	if (attr->ia_valid & ATTR_ATIME) {
		sattr.atime.seconds = attr->ia_atime;
		sattr.atime.useconds = 0;
	}

	error = nfs_proc_setattr(NFS_SERVER(inode), NFS_FH(inode),
		&sattr, &fattr);
	if (!error) {
		nfs_truncate_dirty_pages(inode, sattr.size);
		nfs_refresh_inode(inode, &fattr);
	}
	inode->i_dirt = 0;
	return error;
}

/*
 * Externally visible revalidation function
 */
int
nfs_revalidate(struct inode *inode)
{
	return nfs_revalidate_inode(NFS_SERVER(inode), inode);
}

/*
 * This function is called whenever some part of NFS notices that
 * the cached attributes have to be refreshed.
 *
 * This is a bit tricky because we have to make sure all dirty pages
 * have been sent off to the server before calling invalidate_inode_pages.
 * To make sure no other process adds more write requests while we try
 * our best to flush them, we make them sleep during the attribute refresh.
 *
 * A very similar scenario holds for the dir cache.
 *
 * e.g. nfs file reads call this, so this can invalidate pages if the
 * inode has been changed on the server (so we don't use the in-memory
 * buffer).  We also want to be sure not to use the on-disk buffer,
 * but that is handled at the page-level by nfs_cache_valid_for(inode *, page *)
 */
int
_nfs_revalidate_inode(struct nfs_server *server, struct inode *inode)
{
	struct nfs_fattr fattr;
	int		 status;

	/* this case gets handled by the inline version of the function,
		 nfs_revalidate_inode; (would be redundant iff we know for sure
		 that all calls to this are through the inline fn) */
	if (jiffies - NFS_READTIME(inode) < NFS_ATTRTIMEO(inode))
		return 0;

	dfprintk(PAGECACHE,
					 "NFS: revalidating %x/%ld inode, jiffies -READTIME = %ld, ATTRTIMEO = %ld\n",
					 inode->i_dev, inode->i_ino,jiffies-NFS_READTIME(inode),NFS_ATTRTIMEO(inode));
	/*	NFS_READTIME(inode) = jiffies;  FIXGJB: this gets done in nfs_refresh_inode now */
	if ((status = nfs_proc_getattr(server, NFS_FH(inode), &fattr)) < 0) {
		/*  we try again, just in case FIXGJB, this should really only
				be handled at the RPC level */
		nfs_printk("NFS: _nfs_revalidate_inode: getattr returned %d, trying again\n",status);
		if ((status = nfs_proc_getattr(server, NFS_FH(inode), &fattr)) < 0) {
			nfs_printk("NFS: _nfs_revalidate_inode: getattr returned %d, failing\n",status);
			goto done;
		}
	}

	nfs_refresh_inode(inode,&fattr);

done:
	dfprintk(PAGECACHE,
					 "NFS: inode %x/%ld revalidation complete (status %d). %ld\n",
					 inode->i_dev, inode->i_ino, status,
					 jiffies - NFS_READTIME(inode));
	return status;
	
}


/*
 * File system information
 */
static struct file_system_type nfs_fs_type = {
	nfs_read_super, "nfs", 0, NULL
};

/* see nfs_proc.c for these functions */
int nfs_proc_get_info(char *buf, char **start, off_t offset, int len, int unused);
int nfs_cache_proc_get_info(char *buf, char **start, off_t offset, int len, int unused);
int nfs_usage_proc_get_info(char *buf, char **start, off_t offset, int len, int unused);
int nfs_mounts_proc_get_info(char *buf, char **start, off_t offset, int len, int unused);

#ifdef 0
struct proc_dir_entry nfs_proc_entry = {
  0, /* dynamic inode */
  3, "nfs", /* length and name */
  S_IFDIR | S_IRUGO | S_IXUGO, /* directory read by anyone */
  1, 0, 0, /* nlinks, owner (root), group (root */
  0, NULL, /* size, inode ops unused */
  NULL, /* no get_info interface */
  NULL, NULL, NULL, NULL
};
#endif /* FIXGJB: making /proc/net/nfs/ directory may be more work than it's worth */

struct proc_dir_entry nfs_profile_proc_entry = {
  0, /* dynamic inode */
  11, "nfs-profile", /* length and name */
  S_IFREG | S_IRUGO, /* regular file, read by anyone */
  1, 0, 0, /* nlinks, owner (root), group (root */
  0, NULL, /* size, inode ops unused */
  nfs_proc_get_info, /* use get_info interface */
  NULL, NULL, NULL, NULL
};

struct proc_dir_entry nfs_usage_proc_entry = {
  0, /* dynamic inode */
  9, "nfs-usage", /* length and name */
  S_IFREG | S_IRUGO, /* regular file, read by anyone */
  1, 0, 0, /* nlinks, owner (root), group (root */
  0, NULL, /* size, inode ops unused */
  nfs_usage_proc_get_info, /* use get_info interface */
  NULL, NULL, NULL, NULL
};

struct proc_dir_entry nfs_mounts_proc_entry = {
  0, /* dynamic inode */
  10, "nfs-mounts", /* length and name */
  S_IFREG | S_IRUGO, /* regular file, read by anyone */
  1, 0, 0, /* nlinks, owner (root), group (root */
  0, NULL, /* size, inode ops unused */
  nfs_mounts_proc_get_info, /* use get_info interface */
  NULL, NULL, NULL, NULL
};

struct proc_dir_entry nfs_cache_proc_entry = {
	0, 9, "nfs-cache",
	S_IFREG | S_IRUGO, 1, 0, 0,
/*	0, &proc_net_inode_operations, For when using proc_net_register */
	0, NULL,
	nfs_cache_proc_get_info,
	NULL, NULL, NULL, NULL
};

void
nfs_register_proc_subsystem(void)
{
 	/* now register our proc filesystem "/proc/net/nfs-cache" entry */
/* 	proc_net_register(&nfs_profile_proc_entry); */
 	proc_register_dynamic(&proc_net,&nfs_cache_proc_entry);

 	/* now register our proc filesystem "/proc/net/nfs-profile" entry */
 	proc_register_dynamic(&proc_net,&nfs_profile_proc_entry);
 	/* now register our proc filesystem "/proc/net/nfs-usage" entry */
 	proc_register_dynamic(&proc_net,&nfs_usage_proc_entry);
 	/* now register our proc filesystem "/proc/net/nfs-mounts" entry */
 	proc_register_dynamic(&proc_net,&nfs_mounts_proc_entry);
}

void
nfs_unregister_proc_subsystem(void)
{
/*	proc_net_unregister(nfs_cache_proc_entry.low_ino); */

 	proc_unregister(&proc_net,nfs_cache_proc_entry.low_ino);
 	proc_unregister(&proc_net,nfs_mounts_proc_entry.low_ino);
 	proc_unregister(&proc_net,nfs_usage_proc_entry.low_ino);
 	proc_unregister(&proc_net,nfs_profile_proc_entry.low_ino);
}


static struct file_operations nfs_cache_space_notifier = {
	NULL, /* lseek */
	nfs_read_space, /* read */
  nfs_update_space, /* write */
  NULL, /* readdir */
  NULL, /* select */
  NULL, /* ioctl */
  NULL, /* open */
  NULL, /* release */
  NULL, /* fsync */
  NULL, /* fasync */
  NULL, /* check_media_change */
  NULL, /* revalidate */
};

int
nfs_fillin_daemon(void *dummy)
{
	int ret;
  MOD_INC_USE_COUNT;
	exit_mm(current);
	current->session = 1;
	current->pgrp = 1;
	sprintf(current->comm, "nfsfillind");
	nfs_printk("NFS: nfs_fillin_daemon\n");
	ret = nfs_fillind();
	MOD_DEC_USE_COUNT;
	return ret;
}

/*
 * Initialize NFS
 */
int
init_nfs_fs(void)
{
#ifdef CONFIG_PROC_FS
	rpcstat_register(&nfs_rpcstat);
#endif
#ifndef MODULE
	nfs_register_proc_subsystem();
#ifdef NFS_USE_FILLIND
	kernel_thread(nfs_fillin_daemon,NULL,0);
#endif
	nfs_lookup_cache_allocate();
	if ((nfs_cache_space_major = register(chrdev(0,"nfs-space", &nfs_cache_space_notifier))) != 0)
		printk("NFS: init_module, unable to get major device number for nfs_cache_space_notifier\n");
#endif
	return register_filesystem(&nfs_fs_type);
}

/*
 * Every kernel module contains stuff like this.
 */
#ifdef MODULE
int errno;

/* FIXGJB: what macro has the version? would prefer
  #if KERNEL_MAJOR_VERSION == 2 && KERNEL_MINOR_VERSION >= 1
 or something like that */
#ifdef KERNEL_2_1
EXPORT_NO_SYMBOLS;
/* Not quite true; I just maintain it */
MODULE_AUTHOR("Olaf Kirch <okir@monad.swb.de>");
#endif

extern int nfs_running;

int
init_module(void)
{
	int status;

	if ((status = init_nfs_fs()) == 0)
 		{
		/* register_symtab(0); */
 		/* now register our proc filesystem "/proc/net/nfs" entry */
		nfs_register_proc_subsystem();
		nfs_running = 1;
		kernel_thread(nfs_fillin_daemon,NULL,0);
		nfs_lookup_cache_allocate();
		if (register_chrdev(NFS_SPACE_MAJOR,"nfs-space", &nfs_cache_space_notifier) != 0)
			printk("NFS: init_module, unable to get major device number %d for nfs_cache_space_notifier\n",
						 NFS_SPACE_MAJOR);
 		}
	return status;
}

void
cleanup_module(void)
{
#ifdef CONFIG_PROC_FS
	rpcstat_unregister(&nfs_rpcstat);
#endif
	nfs_running = 0; /* stop the fillind process */
	nfs_unregister_proc_subsystem();
	unregister_filesystem(&nfs_fs_type);
	unregister_chrdev(NFS_SPACE_MAJOR,"nfs-space");
	nfs_lookup_cache_free();
	nfs_free_dircache();
}
#endif
