/* $Id: nfs_cache.h,v 1.23 1997/07/10 16:37:53 gjb Exp $
 * nfs_cache.h
 * (C) 1997 Greg J. Badros -- All Rights Reserved
 * <gjb@cs.washington.edu> 
 * Seattle, WA  USA
 * http://www.cs.washington.edu/homes/gjb
 */

#ifndef _NFS_CACHE_H_
#define _NFS_CACHE_H_
#include "profile.h"
#include "params.h"

/* was for 3 numbers + 1 char separators +
	 server + disk cache root path  --
	 only using two numbers below, though
	 --gjb */
#define NFS_CACHEFILE_MAXLEN 81

/* use the u+x attribute bit on a cache file to indicate a whole file
   is cached */
#define NFS_COMPLETED_BIT S_IXUSR

/* use the g+x attribute bit on a cache file to indicate no more
   caching for that file */
#define NFS_NO_CACHE_BIT S_IXGRP

/* use the u+x attribute bit on the full state file to indicate the
   cache directory doesn't have enough free space to continue caching */
#define NFS_MIN_FREE_VIOL_BIT S_IXUSR

/* use the g+x attribute to indicate that a mounted
   file system is exceeding it's allotted usage
   of the cache */
#define NFS_MAX_SPACE_VIOL_BIT S_IXGRP

/* use g+x to indicate an error when writing to the
	 cache */
#define NFS_ERROR_WRITING_BIT S_IXOTH

/* Major device number for the cache space notifier;
   need to make this device under dev */
#define NFS_SPACE_MAJOR 119

#define NFS_CACHE_UID (0)
#define NFS_CACHE_GID (0)
#define NFS_CACHE_MODE (0600) /* FIXGJB */

#define NFS_CACHE_STATUS_MODE (0600)

/* iput inodeCache when inode is iput if it's got less than
	 this percentages of pages read (or if it's only got
	 1 page read and it's less than 100/this) ) */
#define NFS_NOT_WORTH_KEEPING_AROUND_PCT 10

/* after how many updates do we check disk free space w/ stat */
/* can set this to zero and force user level program to check 
	 for disk full condition (and set +x on .FULL when it is full
	 and cache update should pause) */
/* make this lower for testing -- 256 means 256*4kb=1MB granularity on fullness */
#define NFS_UPDATES_BETWEEN_STATS 256

/* NFS_CACHE_IGNORE_SECONDS is the number of seconds after
   modification that a file will not be cached --
	 it does not rely on synchronized times between client and server--
	 instead, it uses the inode->i_atime as an approximation of the
	 server's time, since we will have just read a page from that inode */
#define NFS_CACHE_IGNORE_SECONDS 30

/* NFS_CACHE_ATIME_IGNORE_SECONDS is the number of seconds of granularity
	 the server's atime will get when we're servicing from our local cache */
#define NFS_CACHE_ATIME_IGNORE_SECONDS 60

#define NFS_FILLED_DUE_MIN_FREE_VIOL  0
#define NFS_FILLED_DUE_MAX_SPACE_VIOL 1
#define NFS_FILLED_DUE_TO_WRITE_ERROR 2

void 
nfs_update_read_cache(struct inode *inode,
		      unsigned long offset,
		      char *buffer,
		      unsigned int countAsked,
		      int result,
		      int fSynch); /* fSynch gratuitous for debugging */


int /* bool */
nfs_cache_valid_for(struct inode *inode, struct page *page);

int /* error */
nfs_cache_readpage(struct inode *inode, struct page *page);

int /* error */
nfs_unlink_cache_file(struct inode *inode);


static inline
int
nfs_is_completely_cached_file(struct inode *inodeCache)
{
#ifdef DEBUG_NFS
  if (inodeCache == NULL) {
		printk("NFS: nfs_is_completely_cached_file: inodeCache == NULL\n");
		return 0;
	}
#endif
	return inodeCache->i_mode & NFS_COMPLETED_BIT;
}

static inline
int
nfs_is_marked_uncacheable_file(struct inode *inodeCache)
{
#ifdef DEBUG_NFS
  if (inodeCache == NULL) {
		printk("NFS: %s: inodeCache == NULL\n",__FUNCTION__);
		return 0;
	}
#endif
	return inodeCache->i_mode & NFS_NO_CACHE_BIT;
}


static inline
int
nfs_is_mostly_uncached_file(struct inode *inodeCache)
{
	struct inode_being_cached_info *pbci;
	int valid;
	int total;
#ifdef DEBUG_NFS
  if (inodeCache == NULL) {
		printk("NFS: nfs_is_mostly_cached_file: inodeCache == NULL\n");
		return 0;
	}
#endif
	pbci = inodeCache->i_ibcinfo;
	valid = pbci->cpagesValid;
	total = pbci->cpagesTotal;
	if (total == 0) {
		printk("NFS: nfs_is_mostly_uncached_file, total == 0!\n");
		return 0;
	}

	if (valid*100/total < NFS_NOT_WORTH_KEEPING_AROUND_PCT)
		return 1;

	if (valid == 1 && total < 100/NFS_NOT_WORTH_KEEPING_AROUND_PCT)
		return 1;

	return 0;
}


int /* error */ nfs_open_cache_file(struct inode *inode, int fNoCreate);

void nfs_remove_incompletely_cached_files(struct super_block *sb);

int /* error */
nfs_set_cache_dir_and_fullstate(struct nfs_server *server,struct nfs_mount_data *data,
				struct nfs_sb_info *nfs_sb);

void nfs_unuse_cache(struct super_block *sb);

void nfs_check_too_full(struct super_block *sb);

int nfs_cleanup_inodes_with_partial_caches(int dev, int fUnlink);

static inline
int /* non-zero is true */
nfs_is_bitmap_valid_for_ipage(const char *const pchBitmap,unsigned long ipage)
{
  int ibyte = ipage >> 3; /* /8 */
  int ibit = ipage - (ibyte << 3); /* *8 */
  int valid = pchBitmap[ibyte] & 1<<(7-ibit);
  return valid;
}

/* check whether we're out of space on the caching partition */
static inline
int
nfs_cache_dir_filled(struct inode *inode)
{
	return NFS_CACHE_FS_INODE(inode)->i_mode & NFS_COMPLETED_BIT;
}

/** For the cache space notifier device */

/* for read Vnode opn */
int nfs_read_space(struct inode *inode, struct file *file, char *buf, int count);
/* for write Vnode opn */
int nfs_update_space(struct inode *inode, struct file *file, const char *buf, int count);

#define NFS_INVALID_MSG(s) do { if (fDoMessage) { printk("NFS: (%d,%ld), page %ld not valid: %s\n", \
																												 inode->i_dev,inode->i_ino, \
																												 page->offset>>PAGE_SHIFT, \
																												 s); } } while (0)
#define NFS_NOT_CACHEABLE_MSG(s) do { if (fDoMessage) { printk("NFS: (%d,%ld), file not cacheable : %s\n", \
																												 inode->i_dev,inode->i_ino, \
																												 s); } } while (0)

static inline
int /* bool */
nfs_is_file_cacheable(struct inode *inode, int iPage)
{
	/* iPage is used only in deciding whether to print an explanation msg of the result */
	int fDoMessage = (iPage == 0);

#ifdef DEBUG_NFS
	if (!inode) {
		printk("NFS: nfs_is_file_cacheable -- inode == NULL!\n");
		return 0;
	}
#endif

	if (!DISK_CACHE_READS(inode) || !DISK_CACHE_ENABLED(inode)) {
		return 0;
	}

	/* FIXGJB: FIXNOW this should be an option! */
	/* do not bother caching files that fit in n page;
		 this may waste an inode spot in the cache --gjb */
	if (inode->i_size <= PAGE_SIZE) {
		verb_nfs_printk("(%d,%ld) too small, not cacheable\n",
							 inode->i_dev,inode->i_ino);
		NFS_NOT_CACHEABLE_MSG("too small");
		return 0;
	}

	/* 4096*4096 = 2^24 = 16MB files */
	if (inode->i_size >= PAGE_SIZE *4096 ) {
		verb_nfs_printk("(%d,%ld) too large, not cacheable\n",
							 inode->i_dev,inode->i_ino);
		NFS_NOT_CACHEABLE_MSG("too large");
		return 0;
	}

	/* do not cache a page that is from a file that has changed
		 very recently (e.g. tar reads the first page, it seems;
		 this uses the access time as an approximation to the servers
		 current time, since we know we've just read a page from that
		 inode */
	if (inode->i_mtime > inode->i_atime - NFS_CACHE_IGNORE_SECONDS) {
		verb_nfs_printk("(%d,%ld) too recently modified (hot), not cacheable\n",
							 inode->i_dev,inode->i_ino);
		/* hot file, would not have cached it already */
		NFS_NOT_CACHEABLE_MSG("too recently modified (hot)");
		return 0;
	}

	if (NFS_CACHE_INODE(inode) && nfs_is_marked_uncacheable_file(NFS_CACHE_INODE(inode))) {
		NFS_NOT_CACHEABLE_MSG("marked uncacheable w/ g+x bit");
		return 0;
	}

	return 1;
}

#endif
