/* $Id: nfs_fillind.c,v 1.18 1997/07/11 23:41:42 gjb Exp $
 * nfs_fillind.c -- provide for filling in the unread parts of files being cached
 * (C) 1997 Greg J. Badros -- All Rights Reserved
 * <gjb@cs.washington.edu> 
 * Seattle, WA  USA
 * http://www.cs.washington.edu/homes/gjb
 */

#define NFS_NEED_XDR_TYPES
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/stat.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/pagemap.h>
#include <linux/sunrpc/clnt.h>
#include <linux/nfs_fs.h>

#include <asm/segment.h>
#include <asm/system.h>

#include "nfs_cache.h"
#include "profile.h"
#include "params.h"

#define fid_printk(x...)  printk( ## x) 

int nfs_running = 0;
int nfs_fillind_paused = 0;

#define SIGNAL_BITMASK(SIG) (1 << ((SIG)-1))

#define NFS_FILLIN_BUSY_STALL_DIVISOR   4 /* fraction of seconds when busy */
#define NFS_FILLIN_IDLE_STALL_MULTIPLIER 5 /* number sec when idle */
#define NFS_FILLIN_MAXRUNS 0

int
nfs_choose_page_to_fillin(struct inode_being_cached_info *pbci)
{
	int ipage = pbci->cpagesTotal;
	char *pchBitmap = pbci->pchValidBitmap;
	while (--ipage >= 0) {
		if (!nfs_is_bitmap_valid_for_ipage(pchBitmap,ipage)) {
			return ipage;
		}
	}
	return -1;
}

/* FIXGJB: this could be asked to fill-in a page
   that exists in a memory buffer;  in that case, the
	 fillin will fail using the high-level read mechanism;
	 I need to do something to either detect this situation,
	 or handle page fillins at a low level.  This happens because
	 *written* pages are in the memory buffer, but never
	 go to the disk cache;  only after the page leaves the
	 memory buffer does this nfs_fillin_page do as intended */
void
nfs_fillin_page(struct inode *inodeCache)
{
	struct inode_being_cached_info *pbci = inodeCache->i_ibcinfo;
	struct inode *inodeNFS = pbci->u.inodeNFS;
	struct file_operations *f_op;
	struct file fileCache;
	int ipage = 0;
	static kdev_t dev_last = 0;
	static unsigned long ino_last = 0;
	static int ipage_last = -1;
	static char szBuffer[PAGE_SIZE+1];
	static int cBytes = 0;

	if (pbci == NULL || pbci->fInodeNFSInMemory == FALSE) {
		fid_printk("NFS: nfs_fillin_page, cannot fill in\n");
	}
	ipage = nfs_choose_page_to_fillin(pbci);
	fid_printk("NFS: %s, asked to fill in (%d,%ld) -> (%d,%ld), page %d\n",
						 __FUNCTION__, inodeNFS->i_dev,inodeNFS->i_ino,
						 inodeCache->i_dev,inodeCache->i_ino,ipage);
	if (ipage <0) {
		printk("NFS: nfs_fillin_page, could not find a valid page -- suspended\n");
		nfs_fillind_paused = 1;
		return;
	}

	inodeNFS->i_count++;
	if (!inodeNFS->i_op) {
		printk("NFS: nfs_fillin_page, i_op == NULL -- suspended\n");
		nfs_fillind_paused = 1;
		goto nfs_fillin_page_done;
	}
	f_op = inodeNFS->i_op->default_file_ops;
	if (f_op == NULL || f_op->write == NULL || f_op->read == NULL) {
		printk("NFS: nfs_fillin_page, f_op{,->{write,read}} == NULL -- suspended\n");
		nfs_fillind_paused = 1;
		goto nfs_fillin_page_done;
	}
	
	fileCache.f_op = f_op;
	fileCache.f_mode = READ;  /* 3 */
	fileCache.f_flags = 0;
	fileCache.f_inode = inodeNFS;
	fileCache.f_pos = ipage * PAGE_SIZE;
	fileCache.f_reada = 0;

	if (ipage == ipage_last &&
			inodeCache->i_ino == ino_last &&
			inodeCache->i_dev == dev_last) {
		/* it's another identical request, so the page must be in a memory buffer, 
			 so we need to update the disk buffer manually from the prior read */
		printk("NFS: nfs_fillind: read %d bytes from memory buffer -- suspended\n",cBytes);
		nfs_update_read_cache(inodeNFS,
													ipage*PAGE_SIZE,
													szBuffer,
													cBytes,
													cBytes,0 /* kernel, not user, space */);
	} else {
		uid_t old_fsuid = current->fsuid;
		current->fsuid = inodeNFS->i_uid; /* "become" owner of file*/
		fileCache.f_count = 1;
		cBytes = f_op->read(inodeNFS,&fileCache,szBuffer,PAGE_SIZE);
		current->fsuid = old_fsuid;
	}

	/* update static variables so we can identify a repeat request */
	dev_last = inodeCache->i_dev;
	ino_last = inodeCache->i_ino;
	ipage_last = ipage;

nfs_fillin_page_done:
	inodeNFS->i_count--;
}

static inline
unsigned long
nfs_score_inodeCache_for_filling(kdev_t dev,unsigned long ino,
																 time_t mtime, time_t atime, 
																 int valid, int total,
																 BOOL fIsInMemory,
																 char *szHostname)
{
	int cpagesLeft = total - valid;
	if (!fIsInMemory)
		return 0;
	if (cpagesLeft <= 0)
		return 0;
/*
	fid_printk("NFS: nfs_score_inodeCache_for_filling: (%d,%ld)\n",
						 dev,ino);*/
	/* shortest job first, favoring larger files */
	/* also, favor files whose mtime is from more than a day ago */
	return 10000000 - cpagesLeft*1000 + total + ((mtime < CURRENT_TIME-86400)?100000:0);
}	

struct inode *
nfs_select_inodeCache_to_fill(void)
{
	struct inode *inode_first = get_first_inode();
	struct inode *inodeCacheBest = NULL;
	struct inode *inodeCache = NULL;
	struct inode_being_cached_info *pbci;
	unsigned long best_score = 0;
	unsigned long score = 0;
	int valid, total;
	BOOL fIsInMemory;
	int i = 0;
	for(inodeCache = inode_first; inodeCache; i++, inodeCache = inodeCache->i_next) {
		/* be sure we do not loop forever! FIXGJB, we'd like the
		 actually # of inodes nr_inodes, but that's not exported now */
		if (i>NR_INODE*8) {
			printk("NFS: nfs_select_inodeCache_to_fill, endless loop\n");
			break;
		}

		if (inodeCache == NULL)
			break;
		if (inodeCache == inode_first && i>0)
			break;
		if (inodeCache->i_count == 0)
			continue;
		if (nfs_is_completely_cached_file(inodeCache))
			continue;
		/* inc count, so we do not reuse the inode (and clear it) from another thread */
		inodeCache->i_count++;
		if ((pbci = inodeCache->i_ibcinfo) != NULL) {
			char *szHostname = "NOSERVER";
			struct inode *inodeNFS;
			if (pbci->nfs_sb_info)
				szHostname = pbci->nfs_sb_info->s_server.hostname;
			valid = pbci->cpagesValid;
			total = pbci->cpagesTotal;
			fIsInMemory = pbci->fInodeNFSInMemory;
			/* for now, only worry about inodes that are in memory already */
			if (fIsInMemory) {
				inodeNFS = pbci->u.inodeNFS;
				if (!inodeNFS) {
					printk("NFS: nfs_select_inodeCache_to_fill, fIsInMemory true, but inodeNFS == NULL (%d,%ld)\n",
								 inodeCache->i_dev,inodeCache->i_ino);
					inodeCache->i_count--;
					continue;
				}
				if (nfs_cache_dir_filled(inodeNFS) ||
						(DC_PAGES_USED_FROM_NFS(inodeNFS) >= DC_MAX_SIZE_FROM_NFS(inodeNFS))) {
					inodeCache->i_count--;
					continue;
				}
				if (nfs_is_file_cacheable(inodeNFS,1 /* fake the page, so we don't get messages */)) {
					/* only score it if it's cacheable -- we do not want to always
						 pick a non-cacheable file as our favorite, and later choose
						 that it's not worth caching */
					score = nfs_score_inodeCache_for_filling(inodeCache->i_dev,
																									 inodeCache->i_ino,
																									 inodeCache->i_mtime,
																									 inodeCache->i_atime,
																									 valid,total,fIsInMemory,szHostname);
/*
					fid_printk("NFS: nfs_select_inodeCache_to_fill (%d,%ld) scored %ld\n",
										 inodeCache->i_dev,inodeCache->i_ino,score);
										 */
					if (NFS_CONGESTED(inodeNFS)) {
						score = 0;
					}

					if (score > best_score) {
						inodeCacheBest = inodeCache;
						best_score = score;
					}
				}
			}
		}
		/* re-allow threads to reuse the inode*/
		inodeCache->i_count--;

	}
	return inodeCacheBest;
}

extern int nfs_async_lookups_enabled;

int nfs_fillind(void)
{
#if NFS_FILLIN_MAXRUNS != 0
	int i=0;
#endif
	static struct inode *inodeCache = NULL;
  fid_printk("NFS: nfs_fillind %d starting\n",current->pid);
  while(nfs_running) {
#if NFS_FILLIN_MAXRUNS != 0
		if (++i > NFS_FILLIN_MAXRUNS) {
			return 0;
		}
#endif
		if (current->signal & SIGNAL_BITMASK(SIGTERM)) {
			return 0;
		}
		if (current->signal & SIGNAL_BITMASK(SIGKILL)) {
			return 0;
		}
		/* Use sigusr1 to clear the profile data out */
		if (current->signal & SIGNAL_BITMASK(SIGUSR1)) {
			printk("NFS: profile counts reset by USR1 signal\n");
			nfs_profile_empty_data();
		}
		/* Use sigusr2 to clear the profile data out */
		if (current->signal & SIGNAL_BITMASK(SIGUSR2)) {
			if (nfs_async_lookups_enabled) {
				nfs_async_lookups_enabled = 0;
				printk("NFS: nfs_async_lookups_enabled reset by USR2 signal\n");
			} else {
				nfs_async_lookups_enabled = 1;
				printk("NFS: nfs_async_lookups_enabled set by USR2 signal\n");
			}
		}
		/* Use sighup to pause and unpause */
		if (current->signal & SIGNAL_BITMASK(SIGHUP)) {
			if (nfs_fillind_paused) {
				nfs_fillind_paused = 0;
				printk("NFS: nfsfillind restarted by HUP signal\n");
			} else {
				nfs_fillind_paused = 1;
				printk("NFS: nfsfillind suspended by HUP signal\n");
			}
		}
		if (nfs_fillind_paused) {
			current->timeout = jiffies + 30 * HZ;
			goto NFS_FILLIND_SLEEP;
		}

		if (inodeCache == NULL)
			inodeCache = nfs_select_inodeCache_to_fill();
		if (inodeCache != NULL)
			nfs_fillin_page(inodeCache);
		if (inodeCache != NULL) {
			/* wait about 1/10th of a second */
			current->timeout = jiffies + HZ/NFS_FILLIN_BUSY_STALL_DIVISOR;
		} else {
			/* nothing currently being filled in,
				 so wait 5 seconds */
			current->timeout = jiffies + NFS_FILLIN_IDLE_STALL_MULTIPLIER * HZ;
		}
	NFS_FILLIND_SLEEP:
		current->signal = 0;
		current->state = TASK_INTERRUPTIBLE;
		schedule();
		inodeCache = NULL; /* reselect a page */
	}
	return 0;
}

